@@ -22,6 +22,7 @@
 
 import os
 import sys
+import json
 import zlib
 import atexit
 import base64
@@ -30,6 +31,7 @@ import hashlib
 import termios
 import subprocess
 
+tmux_output_junk = False
 tmux_real_stdout = sys.stdout
 
 try:
@@ -45,30 +47,16 @@ def reset_stdin_tty():
 
 atexit.register(reset_stdin_tty)
 
-class FileError(Exception):
-    pass
-
-class SendError(Exception):
-    def __init__(self, typ, msg):
+class TrzszError(Exception):
+    def __init__(self, msg, typ=False):
         Exception.__init__(self, msg)
         self.typ = typ
         self.msg = msg
 
     def __str__(self):
         if self.typ:
-            return '[SendError] %s: %s' % (self.typ, self.msg)
-        return '[SendError] %s' % self.msg
-
-class RecvError(Exception):
-    def __init__(self, typ, msg):
-        Exception.__init__(self, msg)
-        self.typ = typ
-        self.msg = msg
-
-    def __str__(self):
-        if self.typ:
-            return '[RecvError] %s: %s' % (self.typ, self.msg)
-        return '[RecvError] %s' % self.msg
+            return '[TrzszError] %s: %s' % (self.typ, self.msg)
+        return '[TrzszError] %s' % self.msg
 
 class Callback(object):
     def on_num(self, num):
@@ -106,8 +94,6 @@ def delay_exit(succ, msg):
         sys.stdout.write(msg + '\n')
     else:
         sys.stderr.write(msg + '\n')
-    if 'TRZSZ_NEW_WINDOW' in os.environ:
-        os.system("""bash -c 'read -s -n 1 -p "Press any key to continue..."'""")
     sys.exit(0 if succ else 1)
 
 def send_line(typ, buf):
@@ -117,8 +103,14 @@ def send_line(typ, buf):
     tmux_real_stdout.write('%s:%s\n' % (typ, enc))
     tmux_real_stdout.flush()
 
-def recv_line(binary=False):
+def recv_line(expect_typ, binary=False, may_has_junk=False):
     s = sys.stdin.readline()
+    if tmux_output_junk or may_has_junk:
+        while s[-2] == '\r':
+            s += sys.stdin.readline()
+        flag = expect_typ + ':'
+        if flag in s:
+            s = flag + s.split(flag)[-1]
     try:
         typ, buf = s.split(':', 1)
         dec = zlib.decompress(base64.b64decode(buf))
@@ -126,18 +118,12 @@ def recv_line(binary=False):
     except (ValueError, TypeError, zlib.error):
         return False, s
 
-def check_succ(ignore_status_bar=False):
-    typ, buf = recv_line()
+def check_succ():
+    typ, buf = recv_line('SUCC')
     if typ == '#EXIT':
         delay_exit(False, buf)
-    if ignore_status_bar and typ != 'SUCC':
-        s = buf if not typ else (typ + ':' + buf)
-        while s[-2] == '\r':
-            s += sys.stdin.readline()
-        if 'SUCC:' in s:
-            return s.split('SUCC:')[-1]
     if typ != 'SUCC':
-        raise SendError(typ, buf)
+        raise TrzszError(buf, typ)
     return buf
 
 def send_succ(info):
@@ -149,66 +135,75 @@ def send_fail(info):
 def send_exit(succ, msg):
     clean_input(0.2)
     send_line('#EXIT', msg)
-    if succ:
-        sys.exit(0)
-    else:
-        sys.exit(1)
+    sys.exit(0 if succ else 1)
 
-def check_exit(succ):
-    typ, buf = recv_line()
-    delay_exit(succ and typ == '#EXIT', buf)
+def send_config(quiet=False, overwrite=False):
+    config = {}
+    if quiet:
+        config['quiet'] = True
+    if overwrite:
+        config['overwrite'] = True
+    if tmux_real_stdout != sys.stdout:
+        config['tmux_output_junk'] = True
+    send_succ(json.dumps(config))
+
+def check_config():
+    typ, cfg = recv_line('SUCC', False, True)
+    if typ != 'SUCC':
+        raise TrzszError(cfg, typ)
+    if cfg == 'OK':
+        return {}
+    try:
+        config = json.loads(cfg)
+    except ValueError:
+        raise TrzszError(cfg, 'JSON INVALID')
+    global tmux_output_junk
+    tmux_output_junk = config.get('tmux_output_junk', False)
+    return config
+
+def check_exit():
+    typ, buf = recv_line('#EXIT')
+    delay_exit(typ == '#EXIT', buf)
 
 def send_check(typ, buf):
     send_line(typ, buf)
     return check_succ()
 
 def recv_check(expect_typ, binary=False):
-    typ, buf = recv_line(binary)
+    typ, buf = recv_line(expect_typ, binary)
     if typ == '#EXIT':
         delay_exit(False, buf)
     if typ != expect_typ:
-        raise RecvError(typ, buf)
+        raise TrzszError(buf, typ)
     return buf
 
 def check_path(dest_path):
     if not os.path.isdir(dest_path):
-        raise FileError('Not a directory: %s' % dest_path)
+        raise TrzszError('Not a directory: %s' % dest_path)
     if not os.access(dest_path, os.W_OK):
-        raise FileError('No permission to write: %s' % dest_path)
+        raise TrzszError('No permission to write: %s' % dest_path)
     return True
 
 def check_files(file_list):
     for f in file_list:
         if not os.path.exists(f):
-            raise FileError('No such file: %s' % f)
+            raise TrzszError('No such file: %s' % f)
         if not os.path.isfile(f):
-            raise FileError('Not a regular file: %s' % f)
+            raise TrzszError('Not a regular file: %s' % f)
         if not os.access(f, os.R_OK):
-            raise FileError('No permission to read: %s' % f)
+            raise TrzszError('No permission to read: %s' % f)
     return True
 
 def check_tmux():
     if 'TMUX' not in os.environ:
         return
-    try:
-        tmux_session = int(os.environ['TMUX'].split(',')[2])
-        output = subprocess.check_output(['tmux', 'lsc', '-t', str(tmux_session), '-F', '#{client_tty} ' \
-                                         '#{client_control_mode}'], stderr=subprocess.STDOUT)
-        tmux_client = output.decode('utf8').strip().split('\n')
-        if len(tmux_client) > 1:
-            sys.stderr.write('Too many clients attached to the tmux session\n')
-            subprocess.call(['tmux', 'lsc'])
-            sys.exit(1)
-        tty, mode = tmux_client[0].split()
-    except (IndexError, ValueError, EnvironmentError, subprocess.CalledProcessError):
+    output = subprocess.check_output(['tmux', 'display-message', '-p', '#{client_tty}:#{client_control_mode}'])
+    tmux_tty, control_mode = output.decode('utf8').strip().split(':')
+    if control_mode == '1' or (not tmux_tty.startswith('/')) or (not os.path.exists(tmux_tty)):
         return
-    if not os.path.exists(tty) or mode != '0':
-        return
-    if 'TRZSZ_NEW_WINDOW' in os.environ:
-        global tmux_real_stdout
-        tmux_real_stdout = open(tty, 'w')
-    else:
-        sys.exit(subprocess.call(['tmux', 'new-window', 'TRZSZ_NEW_WINDOW=1 ' + ' '.join(sys.argv)]))
+    global tmux_real_stdout
+    tmux_real_stdout = open(tmux_tty, 'w')
+    sys.stdout.write('\n\x1b[1A\x1b[0J')
 
 def send_files(file_list, callback=None):
     send_check('NUM', str(len(file_list)))
@@ -254,9 +249,9 @@ def get_new_name(path, name):
         new_name = '%s.%d' % (name, i)
         if not os.path.exists(os.path.join(path, new_name)):
             return new_name
-    raise RecvError(False, 'Fail to assign new file name')
+    raise TrzszError('Fail to assign new file name')
 
-def recv_files(dest_path, callback=None):
+def recv_files(dest_path, callback=None, overwrite=False):
     num = int(recv_check('NUM'))
     send_succ(str(num))
     if callback:
@@ -266,7 +261,7 @@ def recv_files(dest_path, callback=None):
 
     for i in range(num):
         name = recv_check('NAME')
-        file_name = get_new_name(dest_path, name)
+        file_name = name if overwrite else get_new_name(dest_path, name)
         send_succ(file_name)
         if callback:
             callback.on_name(name)
@@ -292,7 +287,7 @@ def recv_files(dest_path, callback=None):
         if digest == m.hexdigest():
             send_succ(digest)
         else:
-            raise RecvError(False, 'Check MD5 of %s failed' % name)
+            raise TrzszError('Check MD5 of %s failed' % name)
         if callback:
             callback.on_done(file_name)
         local_list.append(file_name)
